@@ -1296,6 +1296,16 @@ static void shift(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *da
     REPR(del)->pos_funcs.shift(tc, STABLE(del), del, OBJECT_BODY(del), value, kind);
 }
 
+static void oslice(MVMThreadContext *tc, MVMSTable *st, MVMObject *src, void *data, MVMObject *dest, MVMint64 start, MVMint64 end) {
+    MVMP6opaqueREPRData *repr_data = (MVMP6opaqueREPRData *)st->REPR_data;
+    MVMObject *del;
+    if (repr_data->pos_del_slot == -1)
+        die_no_pos_del(tc, st);
+    data = MVM_p6opaque_real_data(tc, data);
+    del = get_obj_at_offset(data, repr_data->attribute_offsets[repr_data->pos_del_slot]);
+    REPR(del)->pos_funcs.slice(tc, STABLE(del), del, OBJECT_BODY(del), dest, start, end);
+}
+
 static void osplice(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data, MVMObject *target_array, MVMint64 offset, MVMuint64 elems) {
     MVMP6opaqueREPRData *repr_data = (MVMP6opaqueREPRData *)st->REPR_data;
     MVMObject *del;
@@ -1502,7 +1512,7 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
     }
     case MVM_OP_bindattr_o:
     case MVM_OP_bindattrs_o: {
-        MVMSpeshFacts *obj_facts = MVM_spesh_get_and_use_facts(tc, g, ins->operands[1]);
+        MVMSpeshFacts *obj_facts = MVM_spesh_get_and_use_facts(tc, g, ins->operands[0]);
         MVMSpeshFacts *ch_facts = MVM_spesh_get_and_use_facts(tc, g, ins->operands[1]);
         MVMString     *name     = spesh_attr_name(tc, g, ins->operands[2], opcode == MVM_OP_bindattrs_o);
         if (name && ch_facts->flags & MVM_SPESH_FACT_KNOWN_TYPE && ch_facts->type
@@ -1521,7 +1531,7 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
     }
     case MVM_OP_bindattr_i:
     case MVM_OP_bindattrs_i: {
-        MVMSpeshFacts *obj_facts = MVM_spesh_get_and_use_facts(tc, g, ins->operands[1]);
+        MVMSpeshFacts *obj_facts = MVM_spesh_get_and_use_facts(tc, g, ins->operands[0]);
         MVMSpeshFacts *ch_facts = MVM_spesh_get_and_use_facts(tc, g, ins->operands[1]);
         MVMString     *name     = spesh_attr_name(tc, g, ins->operands[2], opcode == MVM_OP_bindattrs_i);
         if (name && ch_facts->flags & MVM_SPESH_FACT_KNOWN_TYPE && ch_facts->type
@@ -1544,7 +1554,7 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
     }
     case MVM_OP_bindattr_n:
     case MVM_OP_bindattrs_n: {
-        MVMSpeshFacts *obj_facts = MVM_spesh_get_and_use_facts(tc, g, ins->operands[1]);
+        MVMSpeshFacts *obj_facts = MVM_spesh_get_and_use_facts(tc, g, ins->operands[0]);
         MVMSpeshFacts *ch_facts = MVM_spesh_get_and_use_facts(tc, g, ins->operands[1]);
         MVMString     *name     = spesh_attr_name(tc, g, ins->operands[2], opcode == MVM_OP_bindattrs_n);
         if (name && ch_facts->flags & MVM_SPESH_FACT_KNOWN_TYPE && ch_facts->type
@@ -1567,7 +1577,7 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
     }
     case MVM_OP_bindattr_s:
     case MVM_OP_bindattrs_s: {
-        MVMSpeshFacts *obj_facts = MVM_spesh_get_and_use_facts(tc, g, ins->operands[1]);
+        MVMSpeshFacts *obj_facts = MVM_spesh_get_and_use_facts(tc, g, ins->operands[0]);
         MVMSpeshFacts *ch_facts = MVM_spesh_get_and_use_facts(tc, g, ins->operands[1]);
         MVMString     *name     = spesh_attr_name(tc, g, ins->operands[2], opcode == MVM_OP_bindattrs_s);
         if (name && ch_facts->flags & MVM_SPESH_FACT_KNOWN_TYPE && ch_facts->type
@@ -1626,6 +1636,7 @@ static const MVMREPROps P6opaque_this_repr = {
         pop,
         unshift,
         shift,
+        oslice,
         osplice,
         MVM_REPR_DEFAULT_AT_POS_MULTIDIM,
         MVM_REPR_DEFAULT_BIND_POS_MULTIDIM,
@@ -1682,6 +1693,11 @@ static void dump_p6opaque(MVMThreadContext *tc, MVMObject *obj, int nested) {
         MVMint16 const num_attributes = repr_data->num_attributes;
         MVMint16 cur_attribute = 0;
         MVMP6opaqueNameMap * const name_to_index_mapping = repr_data->name_to_index_mapping;
+        if (!IS_CONCRETE(obj)) {
+            fprintf(stderr, "(%s", MVM_6model_get_debug_name(tc, obj));
+            fprintf(stderr, nested ? ")" : ")\n");
+            return;
+        }
         fprintf(stderr, "%s.new(", MVM_6model_get_debug_name(tc, obj));
         if (name_to_index_mapping != NULL) {
             MVMint16 i;
@@ -1710,10 +1726,22 @@ static void dump_p6opaque(MVMThreadContext *tc, MVMObject *obj, int nested) {
                             }
                         }
                         else {
-                            MVMString * const s = attr_st->REPR->box_funcs.get_str(tc, attr_st, obj, (char *)data + offset);
-                            char * const str = MVM_string_utf8_encode_C_string(tc, s);
-                            fprintf(stderr, "='%s'", str);
-                            MVM_free(str);
+                            if (attr_st->REPR->ID == MVM_REPR_ID_P6str) {
+                                char * const str = MVM_string_utf8_encode_C_string(tc, (MVMString *)get_obj_at_offset(data, offset));
+                                fprintf(stderr, "='%s'", str);
+                                MVM_free(str);
+                            }
+                            else if (attr_st->REPR->ID == MVM_REPR_ID_P6int) {
+                                MVMint64 val = attr_st->REPR->box_funcs.get_int(tc, attr_st, obj, (char *)data + offset);
+                                fprintf(stderr, "=%ld", val);
+                            }
+                            else {
+                                fprintf(stderr, "[%d]=%s", repr_data->attribute_offsets[slot], MVM_6model_get_stable_debug_name(tc, attr_st));
+                            }
+                            /*MVMString * const s = attr_st->REPR->box_funcs.get_str(tc, attr_st, obj, (char *)data + offset);*/
+                            /*char * const str = MVM_string_utf8_encode_C_string(tc, s);*/
+                            /*fprintf(stderr, "='%s'", str);*/
+                            /*MVM_free(str);*/
                         }
                     }
                     if (cur_attribute++ < num_attributes - 1)
