@@ -792,15 +792,96 @@ static void shift(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *da
     exit_single_user(tc, body);
 }
 
+static void copy_elements(MVMThreadContext *tc, MVMObject *src, MVMObject *dest, MVMint64 s_offset, MVMint64 d_offset, MVMint64 elems) {
+    MVMArrayBody     *s_body      = (MVMArrayBody *)OBJECT_BODY(src);
+    MVMArrayBody     *d_body      = (MVMArrayBody *)OBJECT_BODY(dest);
+    MVMArrayREPRData *s_repr_data = REPR(src)->ID == MVM_REPR_ID_VMArray
+                                    ? (MVMArrayREPRData *)STABLE(src)->REPR_data  : NULL;
+    MVMArrayREPRData *d_repr_data = REPR(src)->ID == MVM_REPR_ID_VMArray
+                                    ? (MVMArrayREPRData *)STABLE(dest)->REPR_data : NULL;
+
+    if (elems > 0) {
+        MVMint64  i;
+        MVMuint16 kind;
+        MVMuint8 d_needs_barrier = dest->header.flags & MVM_CF_SECOND_GEN;
+        if (s_repr_data && d_repr_data
+                && s_repr_data->slot_type == d_repr_data->slot_type
+                && s_repr_data->elem_size == d_repr_data->elem_size
+                && (d_repr_data->slot_type != MVM_ARRAY_OBJ || !d_needs_barrier)
+                && d_repr_data->slot_type  != MVM_ARRAY_STR) {
+            /* Optimized for copying from a VMArray with same slot type */
+            MVMint64 s_start = s_body->start;
+            MVMint64 d_start = d_body->start;
+            memcpy( d_body->slots.u8 + (d_start + d_offset) * d_repr_data->elem_size,
+                    s_body->slots.u8  + (s_start + s_offset) * s_repr_data->elem_size,
+                    d_repr_data->elem_size * elems
+            );
+        }
+        else {
+            switch (s_repr_data->slot_type) {
+                case MVM_ARRAY_OBJ:
+                    kind = MVM_reg_obj;
+                    break;
+                case MVM_ARRAY_STR:
+                    kind = MVM_reg_str;
+                    break;
+                case MVM_ARRAY_I64:
+                case MVM_ARRAY_I32:
+                case MVM_ARRAY_I16:
+                case MVM_ARRAY_I8:
+                    kind = MVM_reg_int64;
+                    break;
+                case MVM_ARRAY_N64:
+                case MVM_ARRAY_N32:
+                    kind = MVM_reg_num64;
+                    break;
+                case MVM_ARRAY_U64:
+                case MVM_ARRAY_U32:
+                case MVM_ARRAY_U16:
+                case MVM_ARRAY_U8:
+                    kind = MVM_reg_int64;
+                    break;
+                default:
+                    abort(); /* never reached, silence compiler warnings */
+            }
+            for (i = 0; i < elems; i++) {
+                MVMRegister to_copy;
+                REPR(src)->pos_funcs.at_pos(tc, STABLE(src), src, s_body, s_offset + i, &to_copy, kind);
+                bind_pos(tc, STABLE(dest), dest, d_body, d_offset + i, to_copy, kind);
+            }
+        }
+    }
+}
+
+static void aslice(MVMThreadContext *tc, MVMSTable *st, MVMObject *src, void *data, MVMObject *dest, MVMint64 start, MVMint64 end) {
+    MVMArrayBody     *s_body      = (MVMArrayBody *)data;
+    MVMArrayBody     *d_body      = (MVMArrayBody *)OBJECT_BODY(dest);
+    MVMArrayREPRData *d_repr_data = REPR(dest)->ID == MVM_REPR_ID_VMArray
+                                      ? STABLE(dest)->REPR_data : NULL;
+
+    MVMint64 total_elems = REPR(src)->elems(tc, st, src, s_body);
+    MVMint64 elems;
+
+    start = start < 0 ? total_elems + start : start;
+    end   = end   < 0 ? total_elems + end   : end;
+    if ( end < start || start < 0 || end < 0 || total_elems <= start || total_elems <= end ) {
+        MVM_exception_throw_adhoc(tc, "MVMArray: Slice index out of bounds");
+    }
+
+    elems = end - start + 1;
+    if (d_repr_data) {
+        set_size_internal(tc, d_body, elems, d_repr_data);
+    }
+
+    copy_elements(tc, src, dest, start, 0, elems);
+}
+
 /* This whole splice optimization can be optimized for the case we have two
  * MVMArray representation objects. */
 static void asplice(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data, MVMObject *from, MVMint64 offset, MVMuint64 count) {
     MVMArrayREPRData *repr_data   = (MVMArrayREPRData *)st->REPR_data;
     MVMArrayBody     *body        = (MVMArrayBody *)data;
 
-    const MVMREPROps      *source_repr = REPR(from);
-    MVMArrayREPRData *source_repr_data = source_repr->ID == MVM_REPR_ID_VMArray ? STABLE(from)->REPR_data : NULL;
-
     MVMint64 elems0 = body->elems;
     MVMint64 elems1 = REPR(from)->elems(tc, STABLE(from), from, OBJECT_BODY(from));
     MVMint64 start;
@@ -875,57 +956,9 @@ static void asplice(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *
     }
     exit_single_user(tc, body);
 
+
     /* now copy C<from>'s elements into SELF */
-    if (elems1 > 0) {
-        MVMint64  i;
-        MVMuint16 kind;
-        MVMuint8 needs_barrier = root->header.flags & MVM_CF_SECOND_GEN;
-        if (source_repr_data
-                && repr_data->slot_type == source_repr_data->slot_type
-                && repr_data->elem_size == source_repr_data->elem_size
-                && (repr_data->slot_type != MVM_ARRAY_OBJ || !needs_barrier)
-                && repr_data->slot_type != MVM_ARRAY_STR) {
-            /* Optimized for copying from a VMArray with same slot type */
-            MVMArrayBody     *from_body        = (MVMArrayBody *)OBJECT_BODY(from);
-            start = body->start;
-            memcpy(body->slots.u8 + (start + offset) * repr_data->elem_size,
-                   from_body->slots.u8 + from_body->start * source_repr_data->elem_size,
-                   repr_data->elem_size * elems1);
-        } else {
-            switch (repr_data->slot_type) {
-                case MVM_ARRAY_OBJ:
-                    kind = MVM_reg_obj;
-                    break;
-                case MVM_ARRAY_STR:
-                    kind = MVM_reg_str;
-                    break;
-                case MVM_ARRAY_I64:
-                case MVM_ARRAY_I32:
-                case MVM_ARRAY_I16:
-                case MVM_ARRAY_I8:
-                    kind = MVM_reg_int64;
-                    break;
-                case MVM_ARRAY_N64:
-                case MVM_ARRAY_N32:
-                    kind = MVM_reg_num64;
-                    break;
-                case MVM_ARRAY_U64:
-                case MVM_ARRAY_U32:
-                case MVM_ARRAY_U16:
-                case MVM_ARRAY_U8:
-                    kind = MVM_reg_int64;
-                    break;
-                default:
-                    abort(); /* never reached, silence compiler warnings */
-            }
-            for (i = 0; i < elems1; i++) {
-                MVMRegister to_copy;
-                REPR(from)->pos_funcs.at_pos(tc, STABLE(from), from,
-                    OBJECT_BODY(from), i, &to_copy, kind);
-                bind_pos(tc, st, root, data, offset + i, to_copy, kind);
-            }
-        }
-    }
+    copy_elements(tc, from, root, 0, offset, elems1);
 }
 
 static void at_pos_multidim(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data, MVMint64 num_indices, MVMint64 *indices, MVMRegister *result, MVMuint16 kind) {
@@ -1349,6 +1382,7 @@ static const MVMREPROps VMArray_this_repr = {
         pop,
         unshift,
         shift,
+        aslice,
         asplice,
         at_pos_multidim,
         bind_pos_multidim,
